clear all;
close all;
clc

addpath('MY_LIB')

%%%%%% Tykhonov regularization %%%%%%

%% generate 3D Shepp Logan phantom

[gt,elli]=phantom3d(256);
gt1 = gt./max(gt(:));
[mh,nh,th]=size(gt);

%% decimation factors in each spatial direction

dr = 2;
dc = 2;
ds = 2;

d = dr*dc*ds;
ml = mh/dr; 
nl = nh/dc;
tl = th/ds;
m = ml*nl*tl;
blksz = [ml, nl, tl];

%% Guassian kernel           
H = fspecial3('gaussian',9,3);
[m0,n0,t0]=size(H);

%% pad the blurring kernel
hpad=padarray(H,floor([mh-m0+1,nh-n0+1,th-t0+1]/2),'pre');
hpad=padarray(hpad,round([mh-m0-1,nh-n0-1,th-t0-1]/2),'post');

hp_c=fftshift(hpad);%centre the dc of the padded image

FB=fftn(hp_c);% fourier transform of the padded blurring kernel , this is also the Lambda
FBC=conj(FB);% conjugate of the fourier transform of the padded blurring kernel
F2B=abs(FB).^2;
%%
%we blur the image in in the fourier domain.
Hx=real(ifftn(FB.*fftn(gt)));

%% add Gaussian noise
N= numel(gt);
BSNRdb = 30;
sigman = fro(Hx-mean(Hx(:)))/sqrt(N*10^(BSNRdb/10));

y = Hx + sigman*randn(size(gt));
y1 = y./max(y(:));

FY = fftn(y);
sig2n = sigman^2;

%% blur the image in the Fourier domain and decimate
dec=y(1:dr:end,1:dc:end,1:ds:end);
dec1 = dec./max(dec(:)); 

%% upsample observed image by conjugate transpose decimation operator
decsize = size(dec);
decrup = decsize(1)*dr;
deccup = decsize(2)*dc;
decsup = decsize(3)*ds;
DTy=zeros(decrup,deccup,decsup);
DTy(1:dr:end,1:dc:end,1:ds:end)=dec;

%% interpolation
tic;
[y,x,z]=...
    ndgrid(linspace(1, size(dec,1),decrup),...
           linspace(1, size(dec,2),deccup),...
           linspace(1, size(dec,3),decsup));...
           
decup_im = interp3(dec, x,y,z,'cubic');
decup_im1 = decup_im./max(decup_im(:));
toc;

psnr_in = calc_PSNR(gt,decup_im);
fprintf('interpo_psnr: PSNR = %g\n',psnr_in);

%%

taup_list = 10; 
lam = taup_list*sig2n;
FR1= FBC.*fftn(DTy)+fftn(2*lam*decup_im);
F2D=1;

%% block processing on 3D blocks 
tic;

 x11 = FB.*FR1./F2D;
 fun = @(x) reshape(ufd_1(x,[1],[2,3]),m,1);
 FBR1 = blcthre3d(x11, blksz, fun);
 invW1 = blcthre3d(F2B./F2D, blksz, fun);
 
 invWBR1 = FBR1./(invW1 + (2*lam*d));
 
 fun2 = @(x) x.*invWBR1;
 FCBinvWBR1 = blockproc3(FBC, blksz, fun2);
 
 FX1 = (FR1-FCBinvWBR1)./F2D/(2*lam);
 Xestalg = real(ifftn(FX1));
 Xestalg1 = Xestalg./max(Xestalg(:));
 
toc;


psnr_analy = calc_PSNR(gt, Xestalg);
fprintf('analyblock_psnr: PSNR = %g\n',psnr_analy);

isnr_analytic = isnr_c(decup_im,gt, Xestalg);
fprintf('analyblock_isnr: ISNR = %g\n',isnr_analytic);

%%
figure;
imshow3D(gt1);
title('original image');
colormap gray; axis on

figure;
imshow3D(dec1);
title('decimated_blur image');
colormap gray; axis on

figure;
imshow3D(decup_im1);
title('est by interpo image');
colormap gray; axis on

figure;
imshow3D(Xestalg1);
title('Algo 1');
colormap gray; axis on 